use std::{env, time::Duration};

use rs48_lib::{game::Game, grid_displayer::GridDisplayer, prelude::*};
use termion::{clear, cursor};
use tokio::{
    io::AsyncWriteExt,
    net::{tcp::OwnedWriteHalf, TcpListener},
    time::sleep,
};

#[derive(Debug, Clone)]
struct Config {
    delay: Duration,
}

#[derive(Debug, Clone)]
enum Error {
    Disconnect,
    GameEnd,
}

#[tokio::main]
async fn main() {
    let url = env::var("URL").unwrap_or_else(|_| "localhost:2048".to_string());
    let delay = env::var("DELAY")
        .map(|s| s.parse::<u64>().unwrap())
        .unwrap_or(500);

    let delay = Duration::from_millis(delay);
    let config = Config { delay };
    let server = TcpListener::bind(url).await.unwrap();

    loop {
        let (stream, address) = server.accept().await.unwrap();
        println!("[server] connection from '{address}'");
        let (_, write) = stream.into_split();
        tokio::spawn(game(write, config.clone()));
    }
}

async fn game(mut write: OwnedWriteHalf, config: Config) {
    'session: loop {
        let rules = GameRules::default();
        let mut game = Game::new(rules);
        let mut controller = SimulatedController::new(50, 20);
        let color_seed: u16 = rand::random();

        '_game: loop {
            let result = turn(
                &mut game,
                &mut controller,
                &mut write,
                config.delay,
                color_seed,
            )
            .await;
            match result {
                Ok(()) => (),
                Err(Error::GameEnd) => continue 'session,
                Err(Error::Disconnect) => return,
            }
        }
    }
}

async fn turn(
    game: &mut Game,
    controller: &mut SimulatedController,
    write: &mut OwnedWriteHalf,
    delay: Duration,
    color_seed: u16,
) -> Result<(), Error> {
    let displayer = GridDisplayer::new(color_seed);
    let display = displayer.display(game.get_board());
    let clear = clear::All;
    let goto = cursor::Goto(1, 1);
    let frame = format!("{clear}{goto}{display}");
    write
        .write_all(frame.as_bytes())
        .await
        .map_err(|_| Error::Disconnect)?;

    let movement = controller.next_move(&*game).map_err(|_| Error::GameEnd)?;
    game.turn(movement).map_err(|_| Error::GameEnd)?;
    sleep(delay).await;
    Ok(())
}
